from jax import random

from jaxns.abc import AbstractMarkovSampler, SeedPoint, PreProcessType
from jaxns.initial_state import find_first_true_indices
from jaxns.types import FloatArray, PRNGKey, LivePoints, Sample


class BaseMarkovSampler(AbstractMarkovSampler):
    """
    A sampler that conditions off a known satisfying point, e.g. a seed point.
    """

    def get_seed_point(self, key: PRNGKey, live_points: LivePoints, log_L_constraint: FloatArray) -> SeedPoint:
        """
        Samples a seed point from the live points.

        Args:
            key: PRNGKey
            live_points: the current live point set. All points satisfy the log-L constraint
            log_L_constraint: a log-L constraint to sample within. Note: Currently, redundant because we assume live
                points satisfies the constraint, but in the future, some points may not and this will be used.

        Returns:
            a seed point
        """
        select_mask = live_points.reservoir.log_L > log_L_constraint
        sample_idx = find_first_true_indices(select_mask, N=1)[0]
        sample_idx = random.randint(key, (), minval=0, maxval=live_points.reservoir.log_L.size)
        return SeedPoint(
            U0=live_points.reservoir.point_U[sample_idx],
            log_L0=live_points.reservoir.log_L[sample_idx]
        )

    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, live_points: LivePoints,
                   preprocess_data: PreProcessType) -> Sample:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            log_L_constraint: the constraint to sample within
            live_points: the current live points reservoir
            preprocess_data: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample
        """
        key, seed_key = random.split(key, 2)
        seed_point = self.get_seed_point(key=seed_key, live_points=live_points, log_L_constraint=log_L_constraint)
        return self.get_sample_from_seed(key=key, seed_point=seed_point, log_L_constraint=log_L_constraint,
                                         preprocess_data=preprocess_data)



import logging
from typing import Tuple, NamedTuple, List

from jax import tree_map, numpy as jnp, random, pmap
from jax._src.lax.control_flow import while_loop, scan
from jax._src.lax.parallel import all_gather

from jaxns.abc import PreProcessType, AbstractSampler, AbstractNestedSampler
from jaxns.common import remove_chunk_dim, add_chunk_dim
from jaxns.statistics import analyse_sample_collection
from jaxns.termination import determine_termination
from jaxns.types import NestedSamplerState, Reservoir, LivePoints, TerminationCondition, int_type
from jaxns.types import PRNGKey, IntArray, BoolArray, FloatArray
from jaxns.utils import collect_samples

logger = logging.getLogger('jaxns')

__all__ = [
    'StaticNestedSampling'
]


def _single_live_point_shrink(
        key: PRNGKey,
        live_points: LivePoints,
        log_L_contour: FloatArray,
        preprocess_data: PreProcessType,
        sampler: AbstractSampler) -> Tuple[Reservoir, LivePoints, FloatArray]:
    """
    Run nested sampling to replace an entire live point reservoir via shrinkage.

    Args:
        key: PRNGKey
        live_points: live points
        log_L_contour: the initial log-L contour of live-points.
            Note: this could be computed from the live_points reservoir but would be inefficient to do so.
        preprocess_data: any data needed to shrink the replace the live point reservoir with i.i.d. samples.
        sampler: sampler to use

    Returns:
        dead point reservoir, live points, the final log-L contour of live points
    """

    class CarryType(NamedTuple):
        key: PRNGKey
        live_points: LivePoints
        log_L_contour: FloatArray

    ResultType = Reservoir

    def body(carry: CarryType, unused_X: IntArray) -> Tuple[CarryType, ResultType]:
        del unused_X
        idx_min = jnp.argmin(carry.live_points.reservoir.log_L)
        dead_point: Reservoir = tree_map(lambda x: x[idx_min], carry.live_points.reservoir)
        log_L_dead = dead_point.log_L

        # contour becomes log_L_dead if log_L_dead is not supremum of live-points, else we choose the original
        # constraint of dead point. Note: we are at liberty to choose any log_L level as a contour so long as we
        # can sample within it uniformly.

        next_dead_contour = jnp.min(carry.live_points.reservoir.log_L.at[idx_min].set(jnp.inf))
        on_supremum = jnp.equal(log_L_dead, next_dead_contour)
        log_L_contour = jnp.where(on_supremum, carry.log_L_contour, log_L_dead)

        # replace dead point with a new sample about contour
        key, sample_key = random.split(carry.key, 2)

        sample = sampler.get_sample(
            key=sample_key,
            log_L_constraint=log_L_contour,
            live_points=carry.live_points,
            preprocess_data=preprocess_data
        )

        live_points_reservoir = tree_map(lambda old, update: old.at[idx_min].set(update),
                                         carry.live_points.reservoir, Reservoir(*sample))
        live_points = carry.live_points._replace(reservoir=live_points_reservoir)

        return CarryType(key=key, live_points=live_points, log_L_contour=log_L_contour), dead_point

    init_carry = CarryType(key=key, live_points=live_points, log_L_contour=log_L_contour)
    init_X = live_points.reservoir.log_L
    (_, live_points, log_L_contour), dead_reservoir = scan(body, init_carry, init_X)
    return (dead_reservoir, live_points, log_L_contour)


def _single_thread_ns(key: PRNGKey, state: NestedSamplerState, live_points: LivePoints,
                      termination_cond: TerminationCondition, samplers: List[AbstractSampler]) -> Tuple[
    IntArray, NestedSamplerState, LivePoints]:
    """
    Runs a single thread of static nested sampling, using all-gather to compute stopping after each live-point
    set shrinkage, which is approximately equivalent to one e-fold decrease in enclosed prior volume. This
    continues until a stopping condition is reached. Due to the all-gather this stopping condition is based on
    all the data across all devices, and is the same on all devices.

    Args:
        key: PRNGKey
        state: the initial state
        live_points: the initial live points
        termination_cond: the termination condition
        samplers: the samplers to use in order

    Returns:
        termination_reason, state, live_points
    """
    state = state._replace(key=key)

    class CarryType(NamedTuple):
        done: BoolArray
        termination_reason: IntArray
        state: NestedSamplerState
        live_points: LivePoints
        log_L_contour: FloatArray

    def stopping_cond(state: NestedSamplerState, live_points: LivePoints, term_cond: TerminationCondition) -> Tuple[
        BoolArray, IntArray]:
        # Collect all live points from all devices to compute stopping condition
        all_live_points: LivePoints = remove_chunk_dim(all_gather(live_points, 'i'))
        evidence_calculation, sample_stats = analyse_sample_collection(
            sample_collection=state.sample_collection,
            sorted_collection=True
        )
        done, termination_reason = determine_termination(
            term_cond=term_cond,
            sample_collection=state.sample_collection,
            evidence_calculation=evidence_calculation,
            live_points=all_live_points
        )
        return done, termination_reason

    def build_body(sampler: AbstractSampler, term_cond: TerminationCondition):
        def body(body_state: CarryType) -> CarryType:
            key, sample_key = random.split(body_state.state.key, 2)
            state = body_state.state._replace(key=key)

            preprocess_data = sampler.preprocess(state=state, live_points=body_state.live_points)

            (dead_reservoir, live_points, log_L_contour) = _single_live_point_shrink(
                key=sample_key,
                live_points=body_state.live_points,
                log_L_contour=body_state.log_L_contour,
                preprocess_data=preprocess_data,
                sampler=sampler
            )
            # Collect dead reservoirs from all devices, and merge into state
            all_dead_reservoir: Reservoir = remove_chunk_dim(all_gather(dead_reservoir, 'i'))
            all_live_points: LivePoints = remove_chunk_dim(all_gather(live_points, 'i'))
            new_state = collect_samples(state, all_dead_reservoir)

            done, termination_reason = stopping_cond(state=new_state, live_points=all_live_points,
                                                     term_cond=term_cond)

            return CarryType(done=done, termination_reason=termination_reason, state=new_state,
                             live_points=live_points, log_L_contour=log_L_contour)

        return body

    # We start with live points sorted. The starting contour is the constraint of the minimal sample.
    sort_idx = jnp.lexsort((live_points.reservoir.log_L_constraint, live_points.reservoir.log_L))
    live_points: LivePoints = tree_map(lambda x: x[sort_idx], live_points)
    log_L_contour = live_points.reservoir.log_L_constraint[0]

    carry_state = CarryType(done=jnp.asarray(False, jnp.bool_),
                            termination_reason=jnp.asarray(0, int_type),
                            state=state, live_points=live_points, log_L_contour=log_L_contour)
    for sampler in samplers:
        # Set efficiency termination point
        term_cond = termination_cond._replace(efficiency_threshold=sampler.efficiency_threshold)
        done, termination_reason = stopping_cond(
            state=carry_state.state,
            live_points=carry_state.live_points,
            term_cond=term_cond
        )

        carry_state = carry_state._replace(
            done=done
        )

        carry_state = while_loop(
            lambda body_state: jnp.bitwise_not(body_state.done),
            build_body(sampler=sampler, term_cond=term_cond),
            carry_state
        )

    return carry_state.termination_reason, carry_state.state, carry_state.live_points


class StaticNestedSampling:
    """
    Performs parallel static sampling up an underlying model
    """

    def __init__(self, samplers: List[AbstractSampler], num_live_points: int, num_parallel_samplers: int = 1):
        if num_live_points % num_parallel_samplers != 0:
            raise ValueError(
                f"num_live_points {num_live_points} must divide num_parallel_samplers {num_parallel_samplers}"
            )
        self.num_live_points = num_live_points
        self.num_parallel_samplers = num_parallel_samplers
        self.samplers = samplers

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.num_live_points} live points over {self.num_parallel_samplers} devices) using {'->'.join(map(repr, self.samplers))}"

    def __call__(self, state: NestedSamplerState, live_points: LivePoints, termination_cond: TerminationCondition
                 ) -> Tuple[IntArray, NestedSamplerState, LivePoints]:
        if live_points.reservoir.log_L.size != self.num_live_points:
            raise ValueError(
                f"live points reservoir is the wrong size. "
                f"Got {live_points.reservoir.log_L.size} but expected {self.num_live_points}.")

        # live_points should be randomly shuffled between devices
        chunked_live_points = add_chunk_dim(live_points, self.num_parallel_samplers)

        keys = random.split(state.key, self.num_parallel_samplers)

        parallel_ns = pmap(lambda key, live_points: _single_thread_ns(
            key=key,
            state=state,
            live_points=live_points,
            termination_cond=termination_cond,
            samplers=self.samplers
        ), axis_name='i')

        chunked_termination_reason, chunked_state, chunked_live_points = parallel_ns(keys, chunked_live_points)

        termination_reason, state = tree_map(lambda x: x[0], (chunked_termination_reason, chunked_state))
        live_points = remove_chunk_dim(chunked_live_points)

        return termination_reason, state, live_points



import logging
from typing import Tuple
from uuid import uuid4

import numpy as np
from jax import random, vmap, jit, numpy as jnp

from jaxns.abc import AbstractModel, PriorModelType
from jaxns.prior import parse_prior, compute_log_likelihood, transform, log_prob_prior, prepare_input
from jaxns.types import PRNGKey, FloatArray, float_type, LikelihoodType, UType, XType, LikelihoodInputType

try:
    import haiku as hk
except ImportError:
    print("You must `pip install dm-haiku` first.")
    raise

__all__ = [
    'Model'
]

logger = logging.getLogger('jaxns')


class Model(AbstractModel):
    """
    Represents a Bayesian model in terms of a generative prior, and likelihood function.
    """

    def __init__(self, prior_model: PriorModelType, log_likelihood: LikelihoodType):
        self._id = str(uuid4())
        self.__prior_model = prior_model
        self.__log_likelihood = log_likelihood

    def _prior_model(self) -> PriorModelType:
        return self.__prior_model

    def _log_likelihood(self) -> LikelihoodType:
        return self.__log_likelihood

    def _parsed_prior(self) -> Tuple[UType, XType]:
        return parse_prior(prior_model=self.prior_model)

    def __hash__(self):
        return hash(self._id)

    def sample_U(self, key: PRNGKey) -> FloatArray:
        return random.uniform(key=key, shape=(self.U_ndims,), dtype=float_type)

    def transform(self, U: UType) -> XType:
        return transform(U=U, prior_model=self.prior_model)

    def forward(self, U: UType, allow_nan: bool = False) -> FloatArray:
        return compute_log_likelihood(U=U, prior_model=self.prior_model, log_likelihood=self.log_likelihood,
                                      allow_nan=allow_nan)

    def log_prob_prior(self, U: UType) -> FloatArray:
        return log_prob_prior(U=U, prior_model=self.prior_model)

    def prepare_input(self, U: UType) -> LikelihoodInputType:
        return prepare_input(U=U, prior_model=self.prior_model)

    def sanity_check(self, key: PRNGKey, S: int):
        U = jit(vmap(self.sample_U))(random.split(key, S))
        log_L = jit(vmap(lambda u: self.forward(u, allow_nan=True)))(U)
        logger.info("Sanity check...")
        for _U, _log_L in zip(U, log_L):
            if jnp.isnan(_log_L):
                logger.info(f"Found bad point: {_U} -> {self.transform(_U)}")
        assert not any(np.isnan(log_L))
        logger.info("Sanity check passed")



import numpy as np
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp, vmap, random

from jaxns import StaticNestedSampling, collect_samples
from jaxns.initial_state import init_sample_collection, get_uniform_init_live_points, \
    get_live_points_from_samples, find_first_true_indices
from jaxns.types import NestedSamplerState, SampleCollection, TerminationCondition
from jaxns.likelihood_samplers.uniform_samplers import UniformSampler

tfpd = tfp.distributions


def test_init_sample_collection(basic3_model):
    model = basic3_model
    sample_collection = init_sample_collection(size=10, model=model)
    assert sample_collection.sample_idx == 0
    assert sample_collection.reservoir.log_L.size == 10


def test_get_live_points_from_samples(basic3_model):
    model = basic3_model

    U = jnp.stack(list(
        map(lambda x: x.flatten(), jnp.meshgrid(jnp.linspace(0., 1., 100), jnp.linspace(0., 1., 100), indexing='ij'))),
        axis=1)
    log_L = vmap(model.forward)(U)
    assert jnp.bitwise_not(jnp.any(jnp.isnan(log_L)))

    n = 100
    live_points = get_uniform_init_live_points(random.PRNGKey(43),
                                               num_live_points=n,
                                               model=model)
    init_state = NestedSamplerState(key=random.PRNGKey(42),
                                    sample_collection=SampleCollection(sample_idx=n,
                                                                       reservoir=live_points.reservoir))
    assert np.unique(init_state.sample_collection.reservoir.log_L).size == n
    state, live_points = get_live_points_from_samples(state=init_state,
                                                      log_L_constraint=-jnp.inf,
                                                      num_live_points=n - 1)

    unique = np.unique(live_points.reservoir.log_L)
    assert unique.size == n - 1

    assert np.unique(state.sample_collection.reservoir.log_L).size == 2  # 1 + empty type
    assert state.sample_collection.sample_idx == 1


def test_static_uniform(basic3_model):
    model = basic3_model

    m = 1000
    sample_collection = init_sample_collection(size=m, model=model)

    state = NestedSamplerState(key=random.PRNGKey(42), sample_collection=sample_collection)

    n = 100
    live_points = get_uniform_init_live_points(random.PRNGKey(43),
                                               num_live_points=n,
                                               model=model)
    efficiency_threshold = 0.1
    ns = StaticNestedSampling(
        samplers=[UniformSampler(model=model, efficiency_threshold=efficiency_threshold)],
        num_live_points=n, num_parallel_samplers=1)
    _, state, live_points = ns(state=state, live_points=live_points,
                               termination_cond=TerminationCondition(max_samples=0))
    state = collect_samples(state=state, new_reservoir=live_points.reservoir)
    assert state.sample_collection.sample_idx > 0
    assert jnp.mean(1. / live_points.reservoir.num_likelihood_evaluations) > efficiency_threshold


def test_find_first_true_indices():
    # Usage:
    arr = jnp.array([False, True, False, True, True, False, True, False])
    assert find_first_true_indices(arr, 3).tolist() == [1, 3, 4]
    assert find_first_true_indices(arr, 1).tolist() == [1]



from typing import NamedTuple

from jaxns.types import BoolArray, IntArray
from jaxns.types import PRNGKey, FloatArray
from jax import random, numpy as jnp
from jax.lax import while_loop

from jaxns.likelihood_samplers.multi_ellipsoid.multi_ellipsoid_utils import ellipsoid_clustering, MultEllipsoidState
from jaxns.likelihood_samplers.multi_ellipsoid.multi_ellipsoid_utils import sample_multi_ellipsoid
from jaxns.abc import AbstractRejectionSampler
from jaxns.statistics import compute_num_live_points_from_unit_threads, compute_evidence_no_stats
from jaxns.types import NestedSamplerState, LivePoints, Sample, int_type

__all__ = ['MultiellipsoidalSampler']


class MultiellipsoidalSampler(AbstractRejectionSampler):
    def __init__(self, depth: int, *args, **kwargs):
        self._depth = depth
        super().__init__(*args, **kwargs)

    @property
    def max_num_ellipsoids(self):
        return 2 ** self._depth

    def preprocess(self, state: NestedSamplerState, live_points: LivePoints) -> MultEllipsoidState:
        key, sampler_key = random.split(state.key)

        num_live_points = compute_num_live_points_from_unit_threads(
            log_L_constraints=state.sample_collection.reservoir.log_L_constraint,
            log_L_samples=state.sample_collection.reservoir.log_L,
            num_samples=state.sample_collection.sample_idx,
            sorted_collection=True
        )
        evidence_calculation = compute_evidence_no_stats(sample_collection=state.sample_collection,
                                                         num_live_points=num_live_points)

        return ellipsoid_clustering(
            key=sampler_key,
            points=live_points.reservoir.point_U,
            log_VS=evidence_calculation.log_X_mean,
            max_num_ellipsoids=self.max_num_ellipsoids,
            method='em_gmm'
        )

    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, live_points: LivePoints,
                   preprocess_data: MultEllipsoidState) -> Sample:
        def _sample_multi_ellipsoid(key: PRNGKey) -> FloatArray:
            _, U = sample_multi_ellipsoid(
                key=key,
                mu=preprocess_data.params.mu,
                radii=preprocess_data.params.radii,
                rotation=preprocess_data.params.rotation,
                unit_cube_constraint=True
            )
            return U

        class CarryState(NamedTuple):
            done: BoolArray
            key: PRNGKey
            U: FloatArray
            log_L: FloatArray
            log_L_constraint: FloatArray
            num_likelihood_evals: IntArray

        def body(carry_state: CarryState):
            key, sample_key = random.split(carry_state.key, 2)
            log_L = self.model.forward(U=carry_state.U)
            num_likelihood_evals = carry_state.num_likelihood_evals + jnp.asarray(1, int_type)
            # backoff by one e-fold per attempt after efficiency threshold reached
            log_L_constraint = jnp.where(num_likelihood_evals > 1. / self.efficiency_threshold,
                                         carry_state.log_L_constraint - 0.1, carry_state.log_L_constraint)
            done = log_L > log_L_constraint
            U = jnp.where(done, carry_state.U, _sample_multi_ellipsoid(key=sample_key))
            return CarryState(done=done, key=key, U=U, log_L=log_L, num_likelihood_evals=num_likelihood_evals,
                              log_L_constraint=log_L_constraint)

        key, sample_key = random.split(key, 2)
        init_carry_state = CarryState(done=jnp.asarray(False),
                                      key=key,
                                      U=_sample_multi_ellipsoid(key=sample_key),
                                      log_L=log_L_constraint,
                                      log_L_constraint=log_L_constraint,
                                      num_likelihood_evals=jnp.asarray(0, int_type))

        carry_state = while_loop(lambda s: jnp.bitwise_not(s.done), body, init_carry_state)

        sample = Sample(point_U=carry_state.U,
                        log_L_constraint=carry_state.log_L_constraint,
                        log_L=carry_state.log_L,
                        num_likelihood_evaluations=carry_state.num_likelihood_evals,
                        num_slices=jnp.asarray(0, int_type),
                        iid=jnp.asarray(True, jnp.bool_))
        return sample



import numpy as np


def count_intervals(S, X):
    if not S:
        return np.zeros(len(X), dtype=int)

    a, b = np.array(list(zip(*S)))
    X = np.array(X)

    # Sort a and b for searchsorted
    a = np.sort(a)
    b = np.sort(b)

    # Count the number of 'a' values that are strictly less than x
    count_a = np.searchsorted(a, X, side='left')
    c = []
    for x in X:
        c.append(np.sum(a < x))

    print(count_a)
    print(c)

    # Count the number of 'b' values that are greater than or equal to x
    count_b = len(b) - np.searchsorted(b, X, side='left')
    c = []
    for x in X:
        c.append(np.sum(b >= x))

    print(count_b)
    print(c)

    return np.minimum(count_a, count_b)

def count_intervals(S, X):
    if not S:
        return np.zeros(len(X), dtype=int)

    a, b = np.array(list(zip(*S)))
    X = np.array(X)

    # Count the number of 'a' values that are less than or equal to x
    count_leq_a = np.searchsorted(a, X, side='right')

    # Count the number of 'b' values that are strictly greater than x
    count_gt_b = len(b) - np.searchsorted(b, X, side='left')

    return len(S) - count_leq_a - count_gt_b


def count_intervals_naive(S, X):
    counts = []

    for x in X:
        count = sum(1 for a, b in S if a < x <= b)
        counts.append(count)

    return counts


def test_basic():
    S = {(1, 5), (2, 6), (4, 8)}
    X = [3, 6, 7]
    assert np.array_equal(count_intervals(S, X), count_intervals_naive(S, X))


def test_empty_intervals():
    S = set()
    X = [3, 6, 7]
    assert np.array_equal(count_intervals(S, X), count_intervals_naive(S, X))


def test_empty_values():
    S = {(1, 5), (2, 6), (4, 8)}
    X = []
    assert np.array_equal(count_intervals(S, X), count_intervals_naive(S, X))


def test_single_interval():
    S = {(1, 5)}
    X = [0, 1, 3, 5, 6]
    assert np.array_equal(count_intervals(S, X), count_intervals_naive(S, X))


def test_non_overlapping_intervals():
    S = {(1, 2), (2, 3), (3, 6), (2, 4), (4, 6), (4, 5)}
    X = [1,2,3,4,5,6]
    assert np.array_equal(count_intervals(S, X), count_intervals_naive(S, X))


def test_overlapping_intervals():
    S = {(1, 4), (3, 6), (5, 8)}
    X = [1, 3, 5, 6, 8]
    assert np.array_equal(count_intervals(S, X), count_intervals_naive(S, X))


def test_consecutive_values():
    S = {(1, 3), (2, 4), (3, 5)}
    X = [3]
    assert np.array_equal(count_intervals(S, X), count_intervals_naive(S, X))


def test_no_valid_intervals():
    S = {(1, 2), (3, 4), (5, 6)}
    X = [7, 8, 9]
    assert np.array_equal(count_intervals(S, X), count_intervals_naive(S, X))


def test_interval_edges():
    S = {(1, 5)}
    X = [1, 5]
    assert np.array_equal(count_intervals(S, X), count_intervals_naive(S, X))



from jaxns import TerminationCondition
from jaxns.types import TerminationConditionDisjunction, TerminationConditionConjunction


def test_termination_condition():
    assert isinstance(TerminationCondition() | TerminationCondition(), TerminationConditionDisjunction)
    assert isinstance(TerminationCondition() & TerminationCondition(), TerminationConditionConjunction)



from abc import ABC, abstractmethod
from functools import cached_property
from typing import TypeVar, NamedTuple, Optional, Union, Tuple, Generator, Callable

from jax import numpy as jnp

from jaxns.internals.shapes import tuple_prod
from jaxns.types import FloatArray, float_type, PRNGKey, IntArray, NestedSamplerState, LivePoints, Sample, \
    TerminationCondition, UType, XType, LikelihoodType, LikelihoodInputType, BoolArray

__all__ = [
    'AbstractModel',
    'AbstractSampler',
    'AbstractRejectionSampler',
    'AbstractMarkovSampler',
    'AbstractNestedSampler',
    'PreProcessType',
    'SeedPoint'
]

PreProcessType = TypeVar('PreProcessType')


class SeedPoint(NamedTuple):
    U0: FloatArray
    log_L0: FloatArray


class AbstractPrior(ABC):
    def __init__(self, name: Optional[str] = None):
        self.name = name

    def __repr__(self):
        return f"{self.name if self.name is not None else '*'}\t{self.base_shape} -> {self.shape} {self.dtype}"

    @abstractmethod
    def _dtype(self):
        ...

    @abstractmethod
    def _base_shape(self) -> Tuple[int, ...]:
        ...

    @abstractmethod
    def _shape(self) -> Tuple[int, ...]:
        ...

    @abstractmethod
    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        ...

    @abstractmethod
    def _inverse(self, X) -> FloatArray:
        ...

    @abstractmethod
    def _log_prob(self, X) -> FloatArray:
        ...

    @property
    def dtype(self):
        return self._dtype()

    @property
    def base_shape(self) -> Tuple[int, ...]:
        return self._base_shape()

    @property
    def base_ndims(self):
        return tuple_prod(self.base_shape)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape()

    def forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._forward(U)

    def inverse(self, X) -> FloatArray:
        return self._inverse(X)

    def log_prob(self, X) -> FloatArray:
        log_prob = self._log_prob(X)
        if log_prob.size > 1:
            log_prob = jnp.sum(log_prob)
        if log_prob.shape != ():
            log_prob = log_prob.reshape(())
        return log_prob


PriorModelGen = Generator[AbstractPrior, jnp.ndarray, Tuple[jnp.ndarray, ...]]
PriorModelType = Callable[[], PriorModelGen]


class AbstractModel(ABC):
    """
    Represents a Bayesian model in terms of a generative prior, and likelihood function.
    """

    @abstractmethod
    def _prior_model(self) -> PriorModelType:
        """
        The prior model.
        """
        ...

    @abstractmethod
    def _log_likelihood(self) -> LikelihoodType:
        """
        The log likelihood function.

        Returns:
            log likelihood function
        """
        ...

    @property
    def prior_model(self) -> PriorModelType:
        """
        The prior model.
        """
        return self._prior_model()

    @property
    def log_likelihood(self) -> LikelihoodType:
        """
        The log likelihood function.

        Returns:
            log likelihood function
        """
        return self._log_likelihood()

    @property
    def U_placeholder(self) -> UType:
        """
        A placeholder for U-space sample.
        """
        return self.parsed_prior[0]

    @property
    def X_placeholder(self) -> XType:
        """
        A placeholder for X-space sample.
        """
        return self.parsed_prior[1]

    @property
    def U_ndims(self) -> int:
        """
        The prior dimensionality.
        """
        return self.U_placeholder.size

    @abstractmethod
    def _parsed_prior(self) -> Tuple[UType, XType]:
        """
        The parsed prior.

        Returns:
            U-space sample, X-space sample
        """
        ...

    @cached_property
    def parsed_prior(self) -> Tuple[UType, XType]:
        """
        The parsed prior.

        Returns:
            U-space sample, X-space sample
        """
        return self._parsed_prior()

    @abstractmethod
    def __hash__(self):
        """
        Hash of the model.
        """
        ...

    @abstractmethod
    def sample_U(self, key: PRNGKey) -> FloatArray:
        """
        Sample uniformly from the prior in U-space.

        Args:
            key: PRNGKey

        Returns:
            U-space sample
        """
        ...

    @abstractmethod
    def transform(self, U: UType) -> XType:
        """
        Compute the prior sample.

        Args:
            U: U-space sample

        Returns:
            prior sample
        """
        ...

    @abstractmethod
    def forward(self, U: UType, allow_nan: bool = False) -> FloatArray:
        """
        Compute the log-likelihood.

        Args:
            U: U-space sample
            allow_nan: whether to allow nans in likelihood

        Returns:
            log likelihood at the sample
        """
        ...

    @abstractmethod
    def log_prob_prior(self, U: UType) -> FloatArray:
        """
        Computes the log-probability of the prior.

        Args:
            U: The U-space sample

        Returns:
            the log probability of prior
        """
        ...

    @abstractmethod
    def prepare_input(self, U: UType) -> LikelihoodInputType:
        """
        Prepares the input for the likelihood function.

        Args:
            U: The U-space sample

        Returns:
            the input to the likelihood function
        """
        ...

    @abstractmethod
    def sanity_check(self, key: PRNGKey, S: int):
        """
        Performs a sanity check on the model.

        Args:
            key: PRNGKey
            S: number of samples to check

        Raises:
            AssertionError: if any of the samples are nan.
        """
        ...


class AbstractSampler(ABC):
    def __init__(self, model: AbstractModel, efficiency_threshold: Optional[FloatArray] = None):
        self.model = model
        if efficiency_threshold is None:
            efficiency_threshold = 0.
        if efficiency_threshold < 0. or efficiency_threshold >= 1.:
            raise ValueError(f"{efficiency_threshold} must be in [0., 1.), got {efficiency_threshold}.")
        efficiency_threshold = jnp.asarray(efficiency_threshold, float_type)
        self.efficiency_threshold = efficiency_threshold

    def __repr__(self):
        return self.__class__.__name__

    @abstractmethod
    def preprocess(self, state: NestedSamplerState, live_points: Union[LivePoints, None] = None) -> PreProcessType:
        """
        Produces a data structure that is necessary for sampling to run.
        Typically this is where clustering happens.

        Args:
            state: nested sampler state

        Returns:
            any valid pytree
        """
        ...

    @abstractmethod
    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, live_points: LivePoints,
                   preprocess_data: PreProcessType) -> Sample:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            log_L_constraint: the constraint to sample within
            live_points: the current live points reservoir
            preprocess_data: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample
        """
        ...


class AbstractRejectionSampler(AbstractSampler):
    """
    Samplers that are based on rejection sampling. They usually first-lines of attack, and are stopped once efficiency
    gets too low.
    """
    pass


class AbstractMarkovSampler(AbstractSampler):
    """
    A sampler that conditions off a known satisfying point, e.g. a seed point.
    """

    @abstractmethod
    def get_sample_from_seed(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                             preprocess_data: PreProcessType) -> Sample:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            seed_point: function that gets the next sample from a seed point
            log_L_constraint: the constraint to sample within
            preprocess_data: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample
        """
        ...


class AbstractNestedSampler(ABC):
    @abstractmethod
    def __call__(self, key: PRNGKey, term_cond: TerminationCondition, *,
                 init_state: Optional[NestedSamplerState] = None) -> Tuple[IntArray, NestedSamplerState]:
        """
        Performs approximate nested sampling followed by adaptive refinement.

        Args:
            key: PRNGKey
            term_cond: termination condition
            init_state: optional initial state

        Returns:
            termination reason, and exact state
        """
        ...


class AbstractDistribution(ABC):
    @abstractmethod
    def _dtype(self):
        ...

    @abstractmethod
    def _base_shape(self) -> Tuple[int, ...]:
        ...

    @abstractmethod
    def _shape(self) -> Tuple[int, ...]:
        ...

    @abstractmethod
    def _forward(self, U):
        ...

    @abstractmethod
    def _inverse(self, X):
        ...

    @abstractmethod
    def _log_prob(self, X):
        ...

    @property
    def dtype(self):
        return self._dtype()

    @property
    def base_shape(self) -> Tuple[int, ...]:
        return self._base_shape()

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape()

    def forward(self, U):
        return self._forward(U)

    def inverse(self, X):
        return self._inverse(X)

    def log_prob(self, X):
        return self._log_prob(X)



import jax
import jax.numpy as jnp
import time

if __name__ == '__main__':
    # Simple function that we'll JIT compile
    def slow_function(x):
        return jnp.dot(x, x)


    # JIT compile the function
    jit_slow_function = jax.jit(slow_function)


    def measure_execution_time(func, shape):
        x = jnp.ones(shape)
        start_time = time.time()
        _ = func(x)
        dt = time.time() - start_time
        print(dt, shape)
        return dt


    # Measure time for two different shapes
    shape_1 = (10000,)
    shape_2 = (20000,)

    # First call with shape_1
    time_1_first_call = measure_execution_time(jit_slow_function, shape_1)

    # First call with shape_2
    time_2_first_call = measure_execution_time(jit_slow_function, shape_2)

    # Second call with shape_1
    time_1_second_call = measure_execution_time(jit_slow_function, shape_1)

    assert time_1_first_call > time_1_second_call, "Caching not effective for shape_1!"
    assert time_1_first_call > time_2_first_call, "Expected first call with different shapes to have similar times!"

    print(f"First call with shape {shape_1}: {time_1_first_call:.6f} seconds")
    print(f"First call with shape {shape_2}: {time_2_first_call:.6f} seconds")
    print(f"Second call with shape {shape_1}: {time_1_second_call:.6f} seconds")



from jax import random, numpy as jnp

from jaxns.statistics import compute_evidence_dual, compute_num_live_points_from_unit_threads, compute_evidence, \
    compute_shrinkage_stats
from jaxns.utils import sample_evidence


def test_nested_sampling_basic(basic_run_results):
    log_Z_true, state, results = basic_run_results
    assert jnp.bitwise_not(jnp.isnan(results.log_Z_mean))
    assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=1.75 * results.log_Z_uncert)


def test_sample_evidence(basic_run_results):
    log_Z_true, state, results = basic_run_results
    log_Z_samples = sample_evidence(random.PRNGKey(42),
                                    results.num_live_points_per_sample,
                                    results.log_L_samples, S=1000)
    assert jnp.isclose(results.log_Z_mean, jnp.mean(log_Z_samples), atol=1e-3)
    assert jnp.isclose(results.log_Z_uncert, jnp.std(log_Z_samples), atol=1e-2)


#
# def test_nested_sampling_basic_parallel():
#     def prior_model() -> PriorModelGen:
#         x = yield Prior(tfpd.Uniform(low=0, high=1), name='x')
#         return x
#
#     def log_likelihood(x):
#         return - jnp.sum(x ** 2)
#
#     model = Model(prior_model=prior_model,
#                   log_likelihood=log_likelihood)
#     exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=2,
#                                   max_samples=1000)
#
#     termination_reason, state = exact_ns(random.PRNGKey(42),
#                                          term_cond=TerminationCondition(live_evidence_frac=1e-4))
#     results = exact_ns.to_results(state, termination_reason)
#
#     exact_ns.plot_diagnostics(results)
#     exact_ns.summary(results)
#     exact_ns.plot_cornerplot(results)
#
#     log_Z_true = bruteforce_evidence(model=model, S=200)
#
#     log_Z_samples = sample_evidence(random.PRNGKey(42),
#                                                results.num_live_points_per_sample,
#                                                results.log_L_samples, S=1000)
#
#     assert jnp.isclose(results.log_Z_mean, jnp.mean(log_Z_samples), atol=1e-3)
#     assert jnp.isclose(results.log_Z_uncert, jnp.std(log_Z_samples), atol=1e-3)
#
#     assert jnp.bitwise_not(jnp.isnan(results.log_Z_mean))
#     assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=1.75 * results.log_Z_uncert)

def test_compute_evidence_dual(basic_mvn_run_results):
    true_logZ, state, results = basic_mvn_run_results

    num_live_points = compute_num_live_points_from_unit_threads(
        log_L_samples=state.sample_collection.reservoir.log_L,
        log_L_constraints=state.sample_collection.reservoir.log_L_constraint,
        num_samples=state.sample_collection.sample_idx,
        sorted_collection=True
    )
    evidence_calculation_dual, stats = compute_evidence_dual(
        sample_collection=state.sample_collection,
        num_live_points=num_live_points
    )
    print('Dual', evidence_calculation_dual)
    evidence_calculation, stats = compute_evidence(
        sample_collection=state.sample_collection,
        num_live_points=num_live_points
    )
    print('Primal', evidence_calculation)

    assert jnp.isclose(evidence_calculation_dual.log_Z_mean, evidence_calculation.log_Z_mean, atol=0.05)


def test_nested_sampling_mvn_static(basic_mvn_run_results):
    true_logZ, state, results = basic_mvn_run_results
    actual_log_Z_mean = results.log_Z_mean
    expected_log_Z_mean = true_logZ
    tol = 1.9 * results.log_Z_uncert
    assert jnp.isclose(actual_log_Z_mean, expected_log_Z_mean, atol=tol)


# def test_nested_sampling_mvn_static_adaptive_refinement(basic_mvn_run_adaptive_refinement_results):
#     true_logZ, state, results = basic_mvn_run_adaptive_refinement_results
#     actual_log_Z_mean = results.log_Z_mean
#     expected_log_Z_mean = true_logZ
#     tol = 1.75 * results.log_Z_uncert
#     assert jnp.isclose(actual_log_Z_mean, expected_log_Z_mean, atol=tol)


def test_nested_sampling_mvn_static_multiellipsoid_sampler(multiellipsoidal_mvn_run_results):
    true_logZ, state, results = multiellipsoidal_mvn_run_results
    actual_log_Z_mean = results.log_Z_mean
    expected_log_Z_mean = true_logZ
    tol = 1.75 * results.log_Z_uncert
    assert jnp.isclose(actual_log_Z_mean, expected_log_Z_mean, atol=tol)


# def test_multi_needs_fewer_slices():
#     def log_normal(x, mean, cov):
#         L = jnp.linalg.cholesky(cov)
#         dx = x - mean
#         dx = solve_triangular(L, dx, lower=True)
#         return -0.5 * x.size * jnp.log(2. * jnp.pi) - jnp.sum(jnp.log(jnp.diag(L))) - 0.5 * dx @ dx
#
#     ndims = 4
#     prior_mu = 2 * jnp.ones(ndims)
#     prior_cov = jnp.diag(jnp.ones(ndims)) ** 2
#
#     data_mu = jnp.zeros(ndims)
#     data_cov = jnp.diag(jnp.ones(ndims)) ** 2
#     data_cov = jnp.where(data_cov == 0., 0.95, data_cov)
#
#     true_logZ = log_normal(data_mu, prior_mu, prior_cov + data_cov)
#     # not super happy with this being 1.58 and being off by like 0.1. Probably related to the ESS.
#     post_mu = prior_cov @ jnp.linalg.inv(prior_cov + data_cov) @ data_mu + data_cov @ jnp.linalg.inv(
#         prior_cov + data_cov) @ prior_mu
#
#     print(f"True post mu:{post_mu}")
#     print(f"True log Z: {true_logZ}")
#
#     def prior_model() -> PriorModelGen:
#         x = yield Prior(
#             tfpd.MultivariateNormalTriL(loc=prior_mu, scale_tril=jnp.linalg.cholesky(prior_cov)),
#             name='x')
#         return x
#
#     def log_likelihood(x):
#         return log_normal(x, data_mu, data_cov)
#
#     model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
#
#     model.sanity_check(random.PRNGKey(52), S=100)
#     exact_ns = ExactNestedSampler(model=model, num_live_points=200, num_parallel_samplers=1,
#                                   max_samples=1e4)
#     with Timer():
#         print("Using Multi Dimensional Slice Sampling")
#         termination_reason, state = exact_ns(random.PRNGKey(42),
#                                              term_cond=TerminationCondition(live_evidence_frac=1e-4))
#         results = exact_ns.to_results(state, termination_reason)
#         exact_ns.summary(results)
#         num_slices_multi = results.total_num_slices
#     assert jnp.isclose(results.log_Z_mean, true_logZ, atol=1.75 * results.log_Z_uncert)
#
#     slice_sampler = UniDimSliceSampler(model=model, midpoint_shrink=True)
#     exact_ns = ExactNestedSampler(model=model, num_live_points=200, num_parallel_samplers=1,
#                                   max_samples=1e4, refinement_slice_sampler=slice_sampler)
#     with Timer():
#         print("Using Uni Dimensional Slice Sampling")
#         termination_reason, state = exact_ns(random.PRNGKey(42),
#                                              term_cond=TerminationCondition(live_evidence_frac=1e-4))
#         results = exact_ns.to_results(state, termination_reason)
#         exact_ns.summary(results)
#         num_slices_uni = results.total_num_slices
#     assert jnp.isclose(results.log_Z_mean, true_logZ, atol=1.75 * results.log_Z_uncert)
#     # TODO(Joshuaalbert): broken test, probably means multidim slice sampling is not robust to structure.
#     # assert num_slices_multi < num_slices_uni

def test_nested_sampling_plateau(plateau_run_results):
    log_Z_true, true_log_Z_uncert, state, results = plateau_run_results
    assert jnp.bitwise_not(jnp.isnan(results.log_Z_mean))
    assert jnp.isclose(results.log_Z_mean, 0., atol=1.75 * true_log_Z_uncert)


def test_basic2_correctness(basic2_results):
    log_Z_true, state, results, X_exact = basic2_results
    assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=results.log_Z_uncert * 1.75)


def test_compute_shrinkage_stats(basic2_results):
    log_Z_true, state, results, X_exact = basic2_results

    # ensure there is no bug in control flow. X should be same to controlled evaluation
    log_X_mean, log_X_std = compute_shrinkage_stats(results.num_live_points_per_sample)
    assert jnp.allclose(results.log_X_mean, log_X_mean)


def test_shrinkage(basic2_results):
    log_Z_true, state, results, X_exact = basic2_results

    diff = results.log_X_mean - jnp.log(X_exact)
    diff = jnp.where(jnp.isfinite(diff), diff, jnp.nan)
    # print(jnp.nanstd(diff))
    assert jnp.nanstd(diff) < 0.26

    log_X_mean, log_X_std = compute_shrinkage_stats(results.num_live_points_per_sample)
    rel_diff = jnp.abs(jnp.exp(log_X_mean) - X_exact) / jnp.exp(log_X_std)
    print("Relative shrinkage errors", jnp.percentile(rel_diff, jnp.asarray([50, 75, 90, 95])))
    # assert jnp.all(jnp.percentile(rel_diff, jnp.asarray([50, 75, 90, 95])) < jnp.asarray([0.9, 1.1, 1.4, 1.5]))



import logging
from typing import Tuple

import jax
import numpy as np
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp, jit
from jax import random
from jax import vmap, tree_map
from jax._src.scipy.special import logsumexp, ndtr
from tqdm import tqdm

from jaxns import ExactNestedSampler, TerminationCondition, NestedSamplerResults, UType, XType, PRNGKey, FloatArray, \
    float_type, LikelihoodType, LikelihoodInputType
from jaxns.abc import AbstractModel, PriorModelType
from jaxns.prior import Prior, SingularDistribution

try:
    import haiku as hk
except ImportError:
    print("You must `pip install dm-haiku` first.")
    raise

try:
    import optax
except ImportError:
    print("You must `pip install optax` first.")
    raise

__all__ = [
    'EM',
    'ParametrisedModel',
    'prior_to_parametrised_singular'
]

tfpd = tfp.distributions
tfpk = tfp.math.psd_kernels

logger = logging.getLogger('jaxns')


def prior_to_parametrised_singular(prior: Prior) -> Prior:
    """
    We use a Normal parameter with centre on unit cube, and scale covering the whole cube, as the base representation.

    Args:
        prior: any prior

    Returns:
        A parameter representing the prior.
    """
    name = f"{prior.name}_param"
    init_value = 0.5 * jnp.ones(prior.base_shape, dtype=float_type)
    norm_U_base_param = hk.get_parameter(
        name=name,
        shape=prior.base_shape,
        dtype=float_type,
        init=hk.initializers.Constant(init_value)
    )
    # transform norm_base_param with normal cdf.
    mu = 0.5
    scale = 0.5
    U_base_param = ndtr((norm_U_base_param - mu) / scale)
    param = prior.forward(U_base_param)
    dist = SingularDistribution(value=param, dist=prior.dist)
    return Prior(dist_or_value=dist, name=prior.name)


class ParametrisedModel(AbstractModel):
    """
    A parametrised model, which is a wrapper around a model and its parameters.
    """

    def __init__(self, base_model: AbstractModel, params: hk.MutableParams | None = None):
        """
        Initialise the parametrised model. This means you can use hk.get_parameter anywhere within the prior or
        likelihood definitions.

        Args:
            base_model: The base model to wrap.
            params: The parameters to use. If None, then you must call init_params and set params
                before using the model.
        """
        self.base_model = base_model
        if params is None:
            params = self.init_params(rng=random.PRNGKey(0))
        self._params = params

    def _prior_model(self) -> PriorModelType:
        return self.base_model.prior_model

    def _log_likelihood(self) -> LikelihoodType:
        return self.base_model.log_likelihood

    @property
    def params(self):
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return self._params

    def new(self, params: hk.MutableParams) -> 'ParametrisedModel':
        """
        Create a new parametrised model with the given parameters.

        **This is (and must be) a pure function.**

        Args:
            params: The parameters to use.

        Returns:
            The new parametrised model.
        """
        return ParametrisedModel(base_model=self.base_model, params=params)

    def __hash__(self):
        return self.base_model.__hash__()

    def _parsed_prior(self) -> Tuple[UType, XType]:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return hk.transform(self.base_model._parsed_prior).apply(params=self._params, rng=None)

    def sanity_check(self, key: PRNGKey, S: int):
        U = jit(vmap(self.sample_U))(random.split(key, S))
        log_L = jit(vmap(lambda u: self.forward(u, allow_nan=True)))(U)
        logger.info("Sanity check...")
        for _U, _log_L in zip(U, log_L):
            if jnp.isnan(_log_L):
                logger.info(f"Found bad point: {_U} -> {self.transform(_U)}")
        assert not any(np.isnan(log_L))
        logger.info("Sanity check passed")

    def sample_U(self, key: PRNGKey) -> FloatArray:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return hk.transform(self.base_model.sample_U).apply(params=self._params, rng=None, key=key)

    def transform(self, U: UType) -> XType:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return hk.transform(self.base_model.transform).apply(params=self._params, rng=None, U=U)

    def forward(self, U: UType, allow_nan: bool = False) -> FloatArray:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return hk.transform(self.base_model.forward).apply(params=self._params, rng=None, U=U,
                                                           allow_nan=allow_nan)

    def log_prob_prior(self, U: UType) -> FloatArray:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return hk.transform(self.base_model.log_prob_prior).apply(params=self._params, rng=None, U=U)

    def prepare_input(self, U: UType) -> LikelihoodInputType:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return hk.transform(self.base_model.prepare_input).apply(params=self._params, rng=None, U=U)

    def init_params(self, rng: PRNGKey) -> hk.MutableParams:
        """
        Initialise the parameters of the model.

        Args:
            rng: PRNGkey to initialise the parameters.

        Returns:
            The initialised parameters.
        """

        def log_joint_prob():
            # A pure function that returns the log joint of model.
            U = self.base_model.sample_U(key=random.PRNGKey(0))
            log_L = self.base_model.forward(U=U)
            log_prior_prob = self.base_model.log_prob_prior(U=U)
            return log_L + log_prior_prob

        params = hk.transform(log_joint_prob).init(rng)
        return params


class EM:
    """
    Evidence Maximisation class, that implements the E and M steps. Iteratively computes the evidence and maximises it.
    """

    def __init__(self, model: ParametrisedModel, learning_rate: float = 1e-2, max_num_epochs: int = 100, gtol=1e-4,
                 log_Z_ftol=1., log_Z_atol=1e-4):
        """
        Initialise the EM class.

        Args:
            model: The model to train.
            learning_rate: The learning rate for the M-step.
            max_num_epochs: The maximum number of epochs to train for.
            gtol: The gradient tolerance for the M-step.
            log_Z_ftol: The tolerance for the change in the evidence as function of log_Z_uncert.
            log_Z_atol: The absolute tolerance for the change in the evidence.
        """
        if not isinstance(model, ParametrisedModel):
            raise ValueError("model must be an instance of ParametrisedModel")
        self.model = model
        self.learning_rate = learning_rate
        self.max_num_epochs = max_num_epochs
        self.gtol = gtol
        self.log_Z_ftol = log_Z_ftol
        self.log_Z_atol = log_Z_atol

    def e_step(self, params: hk.MutableParams) -> NestedSamplerResults:
        """
        The E-step is just nested sampling.

        Args:
            params: The parameters to use.

        Returns:
            The nested sampling results.
        """

        # The E-step is just nested sampling
        model = self.model.new(params=params)
        # Create the nested sampler class. In this case without any tuning.
        exact_ns = ExactNestedSampler(model=model, num_live_points=model.U_ndims * 20, max_samples=1e6)

        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-5))
        results = exact_ns.to_results(state, termination_reason)
        # exact_ns.summary(results)
        # exact_ns.plot_diagnostics(results)
        # exact_ns.plot_cornerplot(results)

        return results

    def m_step(self, results: NestedSamplerResults, params: hk.MutableParams) -> hk.MutableParams:
        """
        The M-step is just evidence maximisation.

        Args:
            results: The nested sampling results.
            params: The parameters to use.

        Returns:
            The updated parameters.
        """

        # The M-step is just evidence maximisation

        def neg_log_evidence(params: hk.MutableParams):
            # Compute the log evidence
            model = self.model.new(params=params)
            # To make manageable, we could do chunked_pmap
            log_L_samples = vmap(model.forward)(results.U_samples)
            # We add the log_Z_mean because log_dp_mean is normalised
            log_weights = results.log_dp_mean - results.log_L_samples + results.log_Z_mean
            log_Z = logsumexp(log_L_samples + log_weights)
            return -log_Z

        optimizer = optax.adam(self.learning_rate)
        opt_state = optimizer.init(params)

        @jax.jit
        def train_step(params, opt_state):
            val, grads = jax.value_and_grad(neg_log_evidence)(params)
            updates, new_opt_state = optimizer.update(grads, opt_state)
            new_params = optax.apply_updates(params, updates)
            # Get L2 of grads
            l_oo_grad = tree_map(lambda x: jnp.max(jnp.abs(x)), grads)
            return val, l_oo_grad, new_params, new_opt_state

        p_bar = tqdm(range(self.max_num_epochs), desc="Training Progress", dynamic_ncols=True)
        for epoch in p_bar:
            neg_log_Z, l_oo, params, opt_state = train_step(params, opt_state)
            p_bar.set_description(f"Step {epoch}: log Z = {-neg_log_Z:.4f}, l_oo = {l_oo}")
            p_bar.refresh()
            # Flatten l_oo pytree and get term condition
            l_oo_small = jnp.all(jnp.asarray(jax.tree_leaves(l_oo)) < self.gtol)
            if l_oo_small:
                print(f"Terminating at step {epoch} due to l(inf) small enough.")
                break

        return params

    def train(self, num_steps: int = 10, params: hk.MutableParams | None = None) -> Tuple[
        NestedSamplerResults, hk.MutableParams]:
        """
        Train the model using EM for num_steps.

        Args:
            num_steps: The number of steps to train for, or until convergence.
            params: The initial parameters to use. If None, then the model's params are used.

        Returns:
            The trained parameters.
        """
        if params is None:
            params = self.model.params
        log_Z = -jnp.inf

        # Initialize the progress bar with a description
        p_bar = tqdm(range(num_steps), desc="Processing Steps", dynamic_ncols=True)

        results = None
        for step in p_bar:
            # Execute the e_step
            results = self.e_step(params=params)
            # Update progress bar description
            p_bar.set_description(f"Step {step}: log Z = {results.log_Z_mean:.4f} +- {results.log_Z_uncert:.4f}")
            p_bar.refresh()

            # Check termination condition
            log_Z_change = jnp.abs(results.log_Z_mean - log_Z)
            if log_Z_change < max(self.log_Z_ftol * results.log_Z_uncert, self.log_Z_atol):
                print(f"Convergence achieved at step {step}.")
                break

            # Update log_Z and log_Z_uncert values
            log_Z = results.log_Z_mean

            # Execute the m_step
            params = self.m_step(results, params)

        if results is None:
            raise RuntimeError("No results were computed.")
        return results, params



from time import monotonic_ns

import pylab as plt
import pytest
from jax import numpy as jnp, random
from jax._src.scipy.linalg import solve_triangular
from tensorflow_probability.substrates import jax as tfp

from jaxns import Prior, Model, ExactNestedSampler, TerminationCondition, bruteforce_evidence, \
    ApproximateNestedSampler, sample_evidence, UniformSampler, MultiellipsoidalSampler
from jaxns.abc import PriorModelGen
from jaxns.adaptive_refinement import AdaptiveRefinement

tfpd = tfp.distributions


class Timer:
    def __enter__(self):
        self.t0 = monotonic_ns()

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(f"Time to execute: {(monotonic_ns() - self.t0) / 1e9} seconds.")


@pytest.fixture(scope='package')
def basic_model():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1), name='x')
        return x

    def log_likelihood(x):
        return - jnp.sum(x ** 2)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    return model


@pytest.fixture(scope='package')
def basic_run_results(basic_model):
    model = basic_model

    exact_ns = ExactNestedSampler(model=model,
                                  num_live_points=50,
                                  num_parallel_samplers=1,
                                  max_samples=1000)
    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)

        # exact_ns.plot_diagnostics(results)
        exact_ns.summary(results)
        # exact_ns.plot_cornerplot(results)

    log_Z_true = bruteforce_evidence(model=model, S=200)
    return log_Z_true, state, results


@pytest.fixture(scope='package')
def basic2_results():
    n = 2

    # Prior is uniform in U[0,1]
    # Likelihood is 1 - x**n
    # Z = 1 - 1/n+1

    log_Z_true = jnp.log(1. - 1. / (n + 1))

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    def log_likelihood(x):
        return jnp.log(1. - x ** n)

    def exact_X(L):
        return (1. - L) ** (1. / n)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    ns = ApproximateNestedSampler(
        model=model,
        num_live_points=50,
        num_parallel_samplers=1,
        max_samples=1000
    )

    termination_reason, state = ns(random.PRNGKey(42),
                                   term_cond=TerminationCondition(live_evidence_frac=1e-4))
    results = ns.to_results(state, termination_reason)
    X_exact = exact_X(jnp.exp(results.log_L_samples))
    return log_Z_true, state, results, X_exact


@pytest.fixture(scope='package')
def basic3_model():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z

    def log_likelihood(z):
        return -z ** 2

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    return model


@pytest.fixture(scope='package')
def plateau_run_results():
    def log_likelihood(x):
        return 0.

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    exact_ns = ExactNestedSampler(
        model=model,
        num_live_points=50,
        max_samples=1000
    )

    termination_reason, state = exact_ns(
        random.PRNGKey(42),
        term_cond=TerminationCondition(live_evidence_frac=1e-4)
    )

    results = exact_ns.to_results(state, termination_reason)
    exact_ns.summary(results)
    exact_ns.plot_diagnostics(results)

    log_Z_samples = sample_evidence(random.PRNGKey(42),
                                    results.num_live_points_per_sample,
                                    results.log_L_samples, S=1000)
    plt.hist(log_Z_samples, bins='auto')
    plt.show()

    true_log_Z_uncert = jnp.std(log_Z_samples)
    print(f"true logZ uncert: {true_log_Z_uncert}")
    log_Z_true = 0.
    return log_Z_true, true_log_Z_uncert, state, results


@pytest.fixture(scope='package')
def basic_mvn_model():
    def log_normal(x, mean, cov):
        L = jnp.linalg.cholesky(cov)
        dx = x - mean
        dx = solve_triangular(L, dx, lower=True)
        return -0.5 * x.size * jnp.log(2. * jnp.pi) - jnp.sum(jnp.log(jnp.diag(L))) \
            - 0.5 * dx @ dx

    ndims = 8
    prior_mu = 15 * jnp.ones(ndims)
    prior_cov = jnp.diag(jnp.ones(ndims)) ** 2

    data_mu = jnp.zeros(ndims)
    data_cov = jnp.diag(jnp.ones(ndims)) ** 2
    data_cov = jnp.where(data_cov == 0., 0.99, data_cov)

    true_logZ = log_normal(data_mu, prior_mu, prior_cov + data_cov)
    # not super happy with this being 1.58 and being off by like 0.1. Probably related to the ESS.
    post_mu = prior_cov @ jnp.linalg.inv(prior_cov + data_cov) @ data_mu + data_cov @ jnp.linalg.inv(
        prior_cov + data_cov) @ prior_mu

    print(f"True post mu:{post_mu}")
    print(f"True log Z: {true_logZ}")

    def prior_model() -> PriorModelGen:
        x = yield Prior(
            tfpd.MultivariateNormalTriL(loc=prior_mu, scale_tril=jnp.linalg.cholesky(prior_cov)),
            name='x')
        return x

    def log_likelihood(x):
        return log_normal(x, data_mu, data_cov)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    ns = ApproximateNestedSampler(
        model=model,
        num_live_points=200,
        num_parallel_samplers=1,
        max_samples=40000
    )

    return true_logZ, model, ns


@pytest.fixture(scope='package')
def basic_mvn_run_results(basic_mvn_model):
    true_logZ, model, ns = basic_mvn_model

    with Timer():
        termination_reason, state = ns(random.PRNGKey(42),
                                       term_cond=TerminationCondition(live_evidence_frac=1e-5))
        results = ns.to_results(state, termination_reason)
        ns.summary(results)
    ns.plot_diagnostics(results)
    return true_logZ, state, results


@pytest.fixture(scope='package')
def basic_mvn_run_adaptive_refinement_results(basic_mvn_model, basic_mvn_run_results):
    true_logZ, model, ns = basic_mvn_model
    true_logZ, state, results = basic_mvn_run_results
    ar = AdaptiveRefinement(model=model, patience=20, num_parallel_samplers=1)
    with Timer():
        ar_state = ar(state=state)
        ar_results = ns.to_results(ar_state, 0)
        ns.summary(ar_results)
    ns.plot_diagnostics(ar_results)
    print(jnp.mean(results.num_live_points_per_sample))
    print(jnp.mean(ar_results.num_live_points_per_sample))
    print(jnp.log(jnp.mean(results.num_live_points_per_sample)) - jnp.log(
        jnp.mean(ar_results.num_live_points_per_sample)))
    return true_logZ, state, ar_results


@pytest.fixture(scope='package')
def multiellipsoidal_mvn_run_results(basic_mvn_model):
    true_logZ, model, ns = basic_mvn_model

    # model.sanity_check(random.PRNGKey(42), S=100)
    ns = ApproximateNestedSampler(
        model=model,
        num_live_points=100,
        num_parallel_samplers=1,
        max_samples=40000,
        sampler_chain=[
            UniformSampler(model=model, efficiency_threshold=0.1),
            MultiellipsoidalSampler(model=model, efficiency_threshold=None, depth=0)
        ]
    )
    with Timer():
        termination_reason, state = ns(random.PRNGKey(41),
                                       term_cond=TerminationCondition(live_evidence_frac=1e-5))
        results = ns.to_results(state, termination_reason)
        ns.summary(results)
    ns.plot_diagnostics(results)
    return true_logZ, state, results



from typing import TypeVar

from jax import numpy as jnp, tree_map

T = TypeVar('T')

__all__ = [
    'add_chunk_dim',
    'remove_chunk_dim'
]


def remove_chunk_dim(py_tree: T) -> T:
    """
    Remove the chunk dimension from a pytree

    Args:
        py_tree: pytree to remove chunk dimension from

    Returns:
        pytree with chunk dimension removed
    """

    def _remove_chunk_dim(a):
        shape = list(a.shape)
        shape = [shape[0] * shape[1]] + shape[2:]
        return jnp.reshape(a, shape)

    return tree_map(_remove_chunk_dim, py_tree)


def add_chunk_dim(py_tree: T, chunk_size: int) -> T:
    """
    Add a chunk dimension to a pytree

    Args:
        py_tree: pytree to add chunk dimension to
        chunk_size: size of chunk dimension

    Returns:
        pytree with chunk dimension added
    """

    def _add_chunk_dim(a):
        shape = list(a.shape)
        shape = [chunk_size, shape[0] // chunk_size] + shape[1:]
        return jnp.reshape(a, shape)

    return tree_map(_add_chunk_dim, py_tree)
